import torch
import torch.nn as nn
import torch.nn.functional as F
from util.nerf_renderer import run_network_xyz

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class Embedder:
    """Positional encoding"""

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.create_embedding_fn()
        
    def create_embedding_fn(self):
        embed_fns = []
        d = self.kwargs['input_dims']
        out_dim = 0
        if self.kwargs['include_input']:
            embed_fns.append(lambda x : x)
            out_dim += d
            
        max_freq = self.kwargs['max_freq_log2']
        N_freqs = self.kwargs['num_freqs']
        
        if self.kwargs['log_sampling']:
            freq_bands = 2.**torch.linspace(0., max_freq, steps=N_freqs)
        else:
            freq_bands = torch.linspace(2.**0., 2.**max_freq, steps=N_freqs)
            
        for freq in freq_bands:
            for p_fn in self.kwargs['periodic_fns']:
                embed_fns.append(lambda x, p_fn=p_fn, freq=freq : p_fn(x * freq))
                out_dim += d
                    
        self.embed_fns = embed_fns
        self.out_dim = out_dim
        
    def embed(self, inputs):
        return torch.cat([fn(inputs) for fn in self.embed_fns], -1)


def get_embedder(multires, i=0, input_dim=3):
    if i == -1:
        return nn.Identity(), 3
    
    embed_kwargs = {
                'include_input' : True,
                'input_dims' : input_dim,
                'max_freq_log2' : multires-1,
                'num_freqs' : multires,
                'log_sampling' : True,
                'periodic_fns' : [torch.sin, torch.cos],
    }
    
    embedder_obj = Embedder(**embed_kwargs)
    embed = lambda x, eo=embedder_obj : eo.embed(x)
    return embed, embedder_obj.out_dim


class NeRF(nn.Module):
    def __init__(self, D=8, W=256, input_ch=3, input_ch_views=3, output_ch=4, feature_ch= 8, skips=[4], use_viewdirs=False):
        """NeRF network model"""

        super(NeRF, self).__init__()
        self.D = D
        self.W = W
        self.input_ch = input_ch
        self.input_ch_views = input_ch_views

        self.feature_ch = feature_ch # 8

        self.skips = skips
        self.use_viewdirs = use_viewdirs
        
        self.pts_linears = nn.ModuleList(
            [nn.Linear(input_ch+feature_ch, W)] + [nn.Linear(W, W) if i not in self.skips else nn.Linear(W + input_ch + feature_ch, W) for i in range(D-1)])

        self.views_linears = nn.ModuleList([nn.Linear(input_ch_views + W, W//2)])

        if use_viewdirs:
            self.feature_linear = nn.Linear(W, W)
            self.alpha_linear = nn.Linear(W, 1)
            # self.rgb_linear = nn.Linear(W//2, 3)
            self.rgb_linear = nn.Linear(W//2, 1)
        else:
            self.output_linear = nn.Linear(W, output_ch)

    def forward(self, x, feature):

        # input_pts [B, 3]
        # input_views [B, 3]
        # feature [B, feature_ch]
        
        input_pts, input_views = torch.split(x, [self.input_ch, self.input_ch_views], dim=-1)
        h = torch.cat([input_pts, feature], -1)

        for i, l in enumerate(self.pts_linears):
            h = self.pts_linears[i](h)
            h = F.relu(h)
            if i in self.skips:
                h = torch.cat([input_pts, feature, h], -1)

        if self.use_viewdirs:
            alpha = self.alpha_linear(h)
            feature = self.feature_linear(h)
            h = torch.cat([feature, input_views], -1)

            for i, l in enumerate(self.views_linears):
                h = self.views_linears[i](h)
                h = F.relu(h)

            rgb = self.rgb_linear(h)
            outputs = torch.cat([rgb, alpha], -1)
        else:
            outputs = self.output_linear(h)
        return outputs # (B, 1)



def create_nerf(args):
    """Instantiate NeRF's MLP model"""

    embed_fn, input_ch = get_embedder(args.multires, args.i_embed, input_dim = args.feature_ch)

    input_ch_views = 0
    embeddirs_fn = None
    if args.use_viewdirs:
        embeddirs_fn, input_ch_views = get_embedder(args.multires_views, -1, input_dim = 3)

    output_ch = 2
    skips = [4]

    feature_ch = args.feature_ch
    if args.i_embed != -1:
        feature_ch = input_ch
        input_ch = 3


    model = NeRF(D=args.netdepth, W=args.netwidth,
                input_ch=input_ch, output_ch=output_ch, skips=skips, feature_ch = feature_ch,
                input_ch_views=input_ch_views, use_viewdirs=args.use_viewdirs).to(device)

    grad_vars = list(model.parameters())

    if args.test_mode != 1:
        network_query_fn = lambda inputs, viewdirs, features, network_fn, z_factor, x_size, y_size, x_norm, y_norm : run_network_xyz(inputs, viewdirs, features, network_fn, z_factor, x_size, y_size, x_norm, y_norm,
                                                                    embed_fn=embed_fn,
                                                                    embeddirs_fn=embeddirs_fn,
                                                                    netchunk=args.netchunk)
    elif args.test_mode == 1:
        network_query_fn_test = lambda inputs, viewdirs, features, network_fn, z_factor, x_size, y_size, x_norm, y_norm : run_network_xyz(inputs, viewdirs, features, network_fn, z_factor, x_size, y_size, x_norm, y_norm,
                                                                    embed_fn=embed_fn,
                                                                    embeddirs_fn=embeddirs_fn,
                                                                    netchunk=args.netchunk_test)


    if args.test_mode != 1:
        # Create optimizer
        optimizer = torch.optim.Adam(params=grad_vars, lr=args.lrate, betas=(0.9, 0.999))
        render_kwargs_train = {
            'network_query_fn' : network_query_fn,
            'perturb' : args.perturb,
            'N_samples' : args.N_samples,
            'network_fn' : [model],
            'raw_noise_std' : args.raw_noise_std,
            "lindisp" : args.lindisp,
            "extract_return" : False
        }
        return render_kwargs_train, grad_vars, optimizer
    
    elif args.test_mode ==1:

        # render_kwargs_test = {k : render_kwargs_train[k] for k in render_kwargs_train}
        render_kwargs_test = {
            "network_query_fn" : network_query_fn_test,
            "perturb" : False,
            "N_samples" : args.N_samples_test,
            "network_fn" : [model],
            "raw_noise_std" : 0,
            "lindisp" : args.lindisp,
            "extract_return" : True,
            "retraw" : args.retraw,

        }
        return render_kwargs_test

