import numpy as np
import torch
import torch.nn.functional as F
import torch.quantization
import copy

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DEBUG = False


def get_LF_rays(coords, **kwargs):
    """make rays from light field coordinates"""

    # Coordinates : (B, [v, u, y, x])

    # xy Normlization
    rays_o = coords[:,2:]
    rays_o[:,0] = rays_o[:,0] / kwargs["y_norm"]
    rays_o[:,1] = rays_o[:,1] / kwargs["x_norm"]

    
    z = torch.zeros(coords.shape[0], 1)

    rays_o = torch.cat([rays_o[:,[1,0]], z], -1)

    # uv Normlization
    rays_d = - coords[:,:2]
    rays_d[:,0] = rays_d[:,0] / kwargs["y_norm"]
    rays_d[:,1] = rays_d[:,1] / kwargs["x_norm"]

    z = torch.ones(coords.shape[0], 1)
    rays_d = torch.cat([rays_d[:,[1,0]], z], -1)

    return rays_o, rays_d # (B, 3), (B, 3)


def sample_features(input, features, z_factor, x_size, y_size, x_norm, y_norm):
    """Extract feature in (x,y,z) from feature volume"""
    # input (sample_num, [z,y,x]) -> (1, 1, 1, sample_num, 3) --repeat--> (B, 1, 1,sample_num, 3)
    # feature (B, feature_ch, [z,y,x])

    input = torch.reshape(input, [-1, input.shape[-1]])

    # Normalize z to [-1 1] for grid_sample
    input[...,2] = input[...,2] / z_factor
    input[...,0] = input[...,0] / (x_size / x_norm)
    input[...,1] = input[...,1] / (y_size / y_norm)


    input = input.reshape(1,1,1,-1,3)
    input = input.repeat(features.shape[0],1,1,1,1)

    out = F.grid_sample(features, input, mode = "bilinear", padding_mode="zeros", align_corners = True) # (B, feature_ch, 1, 1, sample_num)

    out = out.reshape(out.shape[0], out.shape[1], out.shape[4])
    sampled_features = out.permute(0,2,1)

    return sampled_features # (B, sample_num, ch)


def batchify_xyz(fn, chunk):
    """Input coordinates and features to MLP by batch to avoid OOM"""

    if chunk is None:
        return fn
    def ret(inputs, features):
        return torch.cat([fn(inputs[i:i+chunk], features[i:i+chunk]) for i in range(0, inputs.shape[0], chunk)], 0)
    return ret


def run_network_xyz(inputs, viewdirs, features, fn, z_factor, x_size, y_size, x_norm, y_norm, embed_fn, embeddirs_fn, netchunk=1024*64):
    """Run the network"""

    sampled_features = sample_features(copy.deepcopy(inputs), features, z_factor, x_size, y_size, x_norm, y_norm) # (B, sample_num, 4)


    inputs = inputs[None,...].repeat(features.shape[0], 1, 1, 1)
    # embedded = embed_fn(inputs)
    embedded = inputs

    if viewdirs is not None:
        input_dirs = viewdirs[:,None].expand([inputs.shape[1], inputs.shape[2], inputs.shape[3]])
        # embedded_dirs = embeddirs_fn(input_dirs)
        embedded_dirs = input_dirs
        embedded_dirs = embedded_dirs[None,...].repeat(features.shape[0], 1, 1, 1)

    embedded = torch.cat([embedded, embedded_dirs], -1)
    sampled_features = embed_fn(sampled_features)
    
    sampled_features_flat = sampled_features.reshape(-1, sampled_features.shape[-1])
    embedded_flat = embedded.reshape(-1, embedded.shape[-1])
    outputs_flat = batchify_xyz(fn, netchunk)(embedded_flat, sampled_features_flat)

    outputs = torch.reshape(outputs_flat, [features.shape[0]] + list(inputs.shape[1:-1]) + [outputs_flat.shape[-1]])

    return outputs # (B, N_rays, sample_num, rgb)


def batchify_rays(rays_flat, features, chunk=1024*32*32, **kwargs):
    """Render rays in smaller minibatches to avoid OOM"""

    all_ret = {}

    for i in range(0, rays_flat.shape[0], chunk):
        ret = render_rays(rays_flat[i:i+chunk], features, **kwargs)
        for k in ret:
            if k not in all_ret:
                all_ret[k] = []
            all_ret[k].append(ret[k])

    all_ret = {k : torch.cat(all_ret[k], 1) for k in all_ret}

    return all_ret


def render(coords, features, chunk=1024*32, use_viewdirs=False, **kwargs):
    """Get color and density from light field coordinates"""

    rays_o, rays_d = get_LF_rays(coords, **kwargs)

    if use_viewdirs:
        viewdirs = rays_d

        viewdirs = viewdirs / torch.norm(viewdirs, dim=-1, keepdim=True)
        viewdirs = torch.reshape(viewdirs, [-1, 3]).float()

    sh = rays_d.shape
    rays_o = torch.reshape(rays_o, [-1, 3]).float()
    rays_d = torch.reshape(rays_d, [-1, 3]).float()

    near, far = kwargs["near"] * torch.ones_like(rays_d[...,:1]), kwargs["far"] * torch.ones_like(rays_d[...,:1])
    rays = torch.cat([rays_o, rays_d, near, far], -1)

    if use_viewdirs:
        rays = torch.cat([rays, viewdirs], -1)

    # Render
    all_ret = batchify_rays(rays, features, chunk, **kwargs)

    for k in all_ret:
        k_sh = [features.shape[0]] + list(sh[:-1]) + list(all_ret[k].shape[2:])
        all_ret[k] = torch.reshape(all_ret[k], k_sh)

    if kwargs["extract_return"]:
        if kwargs["retraw"]:
            k_extract = ['rgb_map', 'depth_map', 'raw']
        else:
            k_extract = ['rgb_map', 'depth_map']
        ret_list = [all_ret[k] for k in k_extract]
        ret_dict = {k : all_ret[k] for k in all_ret if k not in k_extract}

        return ret_list + [ret_dict] 
    else:
        k_extract = ["rgb_map"]
        ret_list = [all_ret[k] for k in k_extract]

        return ret_list


def raw2outputs(raw, z_vals, rays_d, raw_noise_std=0, pytest=False, extract=False):
    """Volume rendering"""

    raw2alpha = lambda raw, dists, act_fn=F.relu: 1.-torch.exp(-act_fn(raw)*dists)

    dists = z_vals[...,1:] - z_vals[...,:-1]
    dists = torch.cat([dists, torch.Tensor([1e10]).expand(dists[...,:1].shape).to(device)], -1)  # [N_rays, N_samples]

    dists = dists * torch.norm(rays_d[...,None,:], dim=-1)

    rgb = torch.sigmoid(raw[...,:1])  # [N_rays, N_samples, 1]
    noise = 0.
    if raw_noise_std > 0.:
        noise = torch.randn(raw[...,1].shape) * raw_noise_std

        if pytest:
            np.random.seed(0)
            noise = np.random.rand(*list(raw[...,1].shape)) * raw_noise_std
            noise = torch.Tensor(noise)
        alpha = raw2alpha(raw[...,1] + noise.to(device), dists)  # [N_rays, N_samples]
    else:
        alpha = raw2alpha(raw[...,1], dists)  # [N_rays, N_samples]

    weights = alpha * torch.cumprod(torch.cat([torch.ones((alpha.shape[0], 1)).to(device), 1.-alpha + 1e-10], -1), -1)[:, :-1]
    rgb_map = torch.sum(weights[...,None] * rgb, -2)  # [N_rays, 3]

    if extract==False:
        return rgb_map        

    depth_map = torch.sum(weights * z_vals, -1)
    disp_map = 1./torch.max(1e-10 * torch.ones_like(depth_map), depth_map / torch.sum(weights, -1))
    acc_map = torch.sum(weights, -1)

    return rgb_map, disp_map, acc_map, weights, depth_map


def render_rays(ray_batch,
                features,
                network_fn,
                network_query_fn,
                N_samples,
                retraw=False,
                lindisp=False,
                perturb=0.,
                raw_noise_std=0.,
                pytest=False,
                extract_return=False,
                near=-1,
                far=1,
                y_norm=1,
                x_norm=1,
                x_size=1,
                y_size=1,
                z_factor=1.):
    """Rendering rays from ray batch"""

    N_rays = ray_batch.shape[0]
    rays_o, rays_d = ray_batch[:,0:3], ray_batch[:,3:6] # [N_rays, 3] each
    viewdirs = ray_batch[:,-3:] if ray_batch.shape[-1] > 8 else None
    bounds = torch.reshape(ray_batch[...,6:8], [-1,1,2])
    near, far = bounds[...,0], bounds[...,1] # [-1,1]

    t_vals = torch.linspace(0., 1., steps=N_samples)
    if not lindisp:
        z_vals = near * (1.-t_vals) + far * (t_vals)
    else:
        z_vals = 1./(1./near * (1.-t_vals) + 1./far * (t_vals))

    z_vals = z_vals.expand([N_rays, N_samples])

    if perturb > 0.:
        # get intervals between samples
        mids = .5 * (z_vals[...,1:] + z_vals[...,:-1])
        upper = torch.cat([mids, z_vals[...,-1:]], -1)
        lower = torch.cat([z_vals[...,:1], mids], -1)
        # stratified samples in those intervals
        t_rand = torch.rand(z_vals.shape)

        # Pytest, overwrite u with numpy's fixed random numbers
        if pytest:
            np.random.seed(0)
            t_rand = np.random.rand(*list(z_vals.shape))
            t_rand = torch.Tensor(t_rand)

        z_vals = lower + (upper - lower) * t_rand


    pts = rays_o[...,None,:] + rays_d[...,None,:] * z_vals[...,:,None] # [N_rays, N_samples, 3]
    pts[...,2] = (pts[...,2]/ far) * z_factor
    raw = network_query_fn(pts.to(device), viewdirs.to(device), features, network_fn[0], z_factor, x_size, y_size, x_norm, y_norm)

    """Expand variables to fit raw output"""
    B = raw.shape[0]
    raw = raw.reshape(raw.shape[0]*raw.shape[1], raw.shape[2], raw.shape[3])
    z_vals = z_vals[None,...].repeat(B, 1,1)
    rays_d = rays_d[None, ...].repeat(B,1,1)
    z_vals = z_vals.reshape(-1, z_vals.shape[-1])
    rays_d = rays_d.reshape(-1, rays_d.shape[-1])


    if extract_return:
        rgb_map, disp_map, acc_map, weights, depth_map = raw2outputs(raw, z_vals.to(device), rays_d.to(device), raw_noise_std, pytest=pytest, extract=True)

        rgb_map = rgb_map.reshape(B, -1, rgb_map.shape[-1])
        disp_map = disp_map.reshape(B, -1, rgb_map.shape[-1])
        acc_map = acc_map.reshape(B, -1, rgb_map.shape[-1])
        depth_map = depth_map.reshape(B, -1, rgb_map.shape[-1])

        ret = {'rgb_map' : rgb_map, "depth_map" : depth_map}

        if retraw:
            ret['raw'] = raw[None,...]
        for k in ret:
            if (torch.isnan(ret[k]).any() or torch.isinf(ret[k]).any()) and DEBUG:
                print(f"! [Numerical Error] {k} contains nan or inf.")
    else:
        rgb_map = raw2outputs(raw, z_vals.to(device), rays_d.to(device), raw_noise_std, pytest=pytest, extract=False)

        rgb_map = rgb_map.reshape(B, -1, rgb_map.shape[-1])

        ret = {"rgb_map" : rgb_map}
        for k in ret:
            if (torch.isnan(ret[k]).any() or torch.isinf(ret[k]).any()) and DEBUG:
                print(f"! [Numerical Error] {k} contains nan or inf.")

    return ret
