from PIL import Image
import matplotlib.pyplot as plt
from model import*
from tqdm import trange
import cv2
import sys
import os

from util.nerf_renderer import *

class CodedCamera():
    def __init__(self, args, save_dir, device):
        self.scene = args.scene
        self.angRes = args.angRes

        self.save_base = save_dir
        self.device = device
        self.save_orig_img = args.save_orig_img
        self.channel_num = args.channel_num
        self.cut_edge = args.cut_edge

        self.shots_num = args.shots_num
        self.pix_mask_size = args.pix_mask_size
        self.pix_mask_num = self.pix_mask_size*self.pix_mask_size
        self.view_point_num = args.angRes*args.angRes

        if args.coded_img_path is None:
            self.coded_img_dir = self.make_savedir("coded_img")
            if self.save_orig_img:
                self.orig_dir = self.make_savedir("original_views")

        if self.cut_edge:
            self.edge_size = 9

    def prepare_data(self, src_path, split):

        def rgb2ycbcr(x):
            y = np.zeros(x.shape, dtype='double')
            y[:,:,0] =  65.481 * x[:, :, 0] + 128.553 * x[:, :, 1] +  24.966 * x[:, :, 2] +  16.0
            y[:,:,1] = -37.797 * x[:, :, 0] -  74.203 * x[:, :, 1] + 112.000 * x[:, :, 2] + 128.0
            y[:,:,2] = 112.000 * x[:, :, 0] -  93.786 * x[:, :, 1] -  18.214 * x[:, :, 2] + 128.0

            y = y / 255.0
            return y

            
        src_datasets = os.listdir(os.path.join(src_path, split))
        src_datasets.sort()
        src_sub_dataset = src_path + split + '/'
        scenes = [self.scene]

        for root, dirs, files in os.walk(src_sub_dataset):
            iters = dirs
            for file in iters:
                if file not in scenes:
                    continue
                idx_scene_save = 0

                imgs = [os.path.join(root+file, f) for f in sorted(os.listdir(root+file))]
                im = cv2.imread(imgs[0])
                H = im.shape[0]
                W = im.shape[1]
                LF = np.zeros((self.angRes,self.angRes,H,W,3))
                
                for img in imgs:
                    im = cv2.imread(img)
                    im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
                    img = img.split("/")[-1]
                    v = int(img[:2])
                    u = int(img[3:5])
                    LF[v,u] = im

                LF = LF / 255.

                (U, V, H, W, _) = LF.shape
                H = H // 8 * 8
                W = W // 8 * 8

                self.H = H
                self.W = W

                # Extract central angRes * angRes views
                LF = LF[(U-self.angRes)//2:(U+self.angRes)//2, (V-self.angRes)//2:(V+self.angRes)//2, 0:H, 0:W, 0:3]
                LF = LF.astype('double')
                (U, V, H, W, _) = LF.shape

                idx_scene_save = idx_scene_save + 1
                LF_y = np.zeros((U*V, H, W),dtype='single')
                LF_r = np.zeros((U*V, H, W),dtype='single')
                LF_g = np.zeros((U*V, H, W),dtype='single')
                LF_b = np.zeros((U*V, H, W),dtype='single')
                LF_rgb = np.zeros((3, U*V, H, W),dtype='single')

                for u in range(U):
                    for v in range(V):
                        tmp_rgb = LF[u, v, :, :, :]
                        LF_r[u*U + v, :, :] = tmp_rgb[:,:,0]
                        LF_g[u*U + v, :, :] = tmp_rgb[:,:,1]
                        LF_b[u*U + v, :, :] = tmp_rgb[:,:,2]
                        LF_rgb[0, u*U + v, :, :] = tmp_rgb[:,:,0]
                        LF_rgb[1, u*U + v, :, :] = tmp_rgb[:,:,1]
                        LF_rgb[2, u*U + v, :, :] = tmp_rgb[:,:,2]

                        tmp_ycbcr = rgb2ycbcr(tmp_rgb)
                        LF_y[u*U + v, :, :] = tmp_ycbcr[:, :, 0]

                        pass
                    pass

                if self.channel_num == 1:
                    LF_data = LF_y
                else:
                    LF_r = LF_r
                    LF_g = LF_g
                    LF_b = LF_b
                    LF_data = np.stack([LF_r, LF_g, LF_b], axis=3)

                return LF_data

        print("Can't load data. Please set correct data path.")
        sys.exit(0)

    def make_savedir(self, name):
        savedir = self.save_base.joinpath(name)
        savedir.mkdir(exist_ok=True)
        return savedir

    def make_original_views(self, data):
        data = data.to(device)
        data[data < 0.0]  = 0.0
        data[data > 1.0]  = 1.0
        if self.channel_num == 1:
                _, self.view_point_num, self.H, self.W = data.shape
                self.channel_num = 1
                data = data.reshape(_, self.view_point_num, self.H, self.W, 1)
        else:
                _, self.view_point_num, self.H, self.W, self.channel_num = data.shape


        if self.save_orig_img:
            orig_lf_data = data[0].to("cpu").numpy()
            if self.cut_edge:
                orig_lf_data = orig_lf_data[...,self.edge_size:-self.edge_size,self.edge_size:-self.edge_size,:]

                self.H_ = self.H - 2 * self.edge_size
                self.W_ = self.W - 2 * self.edge_size
            else:
                self.H_ = self.H
                self.W_ = self.W
                    
            orig_light_field = np.zeros([self.view_point_num, self.H_, self.W_, self.channel_num], np.uint8)
            for t in range(self.view_point_num):
                orig_light_field[t] = np.uint8(orig_lf_data[t]*255.0 + 0.5)

                if self.channel_num == 1:
                    img_save_orig = Image.fromarray(orig_light_field[t,...,0])
                else:
                    img_save_orig = Image.fromarray(orig_light_field[t])

                u = int(t%np.sqrt(self.view_point_num))
                v = int(t/np.sqrt(self.view_point_num))
                img_save_orig.save(str(self.orig_dir) + "/%02d_%02d.png" % (v, u))

            print("Saved 5x5 original views in {}".format(self.orig_dir))

        return data


    def make_coded_img(self, data, aperture_mask, pixel_mask):

        def shot_test(gray_lf_images, aperture_mask, pixel_mask, H, W, pix_mask_size, pix_mask_num, shots_num, view_point_num, dev = 0.005):
                sub_H = int(H/pix_mask_size)
                sub_W = int(W/pix_mask_size)

                # aperture coding
                sub_time_image = torch.cat((aperture_mask[0](gray_lf_images),
                                            aperture_mask[1](gray_lf_images),
                                            aperture_mask[2](gray_lf_images),
                                            aperture_mask[3](gray_lf_images)), dim = 1)


                sub_time_image = sub_time_image.reshape(sub_time_image.shape[0],shots_num,sub_H,pix_mask_size,sub_W,pix_mask_size).permute(0,3,5,1,2,4).reshape(sub_time_image.shape[0],pix_mask_size**2,shots_num,sub_H,sub_W)

                sub_acquired_image = []

                # exposure coding
                for n in range(pix_mask_num):
                        noise = shots_num*view_point_num*dev*torch.randn(1,1,int(H/pix_mask_size),int(W/pix_mask_size)).to("cuda")
                        sub_acquired_image.append(pixel_mask[n](sub_time_image[:,n,:,:,:]) + noise)

                sub_acquired_image = torch.cat(sub_acquired_image[:], dim = 1)

                # Reshape acuired image to (ch, H, W) from (ch, 8, 8, H//8, W//8)
                acquired_image = sub_acquired_image.reshape(sub_acquired_image.shape[0],pix_mask_size,pix_mask_size,sub_H,sub_W).permute(0,3,1,4,2).reshape(sub_acquired_image.shape[0],H,W)

                return acquired_image

        if self.save_orig_img:
            self.make_original_views(data)

        data = data.to(device)
        data[data < 0.0]  = 0.0
        data[data > 1.0]  = 1.0
        if self.channel_num == 1:
                _, self.view_point_num, self.H, self.W = data.shape
                self.channel_num = 1
                data = data.reshape(_, self.view_point_num, self.H, self.W, 1)
        else:
                _, self.view_point_num, self.H, self.W, self.channel_num = data.shape


        coded_img_save = np.zeros([self.H, self.W, self.channel_num], np.float32)
        for ch in range(self.channel_num):

            """Make coded image from original views (5x5)"""
            coded_img = shot_test(data[...,ch], aperture_mask, pixel_mask, self.H, self.W, self.pix_mask_size, self.pix_mask_num, self.shots_num, self.view_point_num)
            coded_img_save[:,:,ch] = coded_img.to("cpu").detach().numpy()


        ''' Save coded image '''
        self.cut_edge = False
        if self.cut_edge:
            cut_coded_img_save = coded_img_save[...,self.edge_size:-self.edge_size,self.edge_size:-self.edge_size,:]
        else:
            cut_coded_img_save = coded_img_save

        img_save = cut_coded_img_save/(self.shots_num*self.view_point_num)

        img_save[img_save < 0.0] = 0.0
        img_save[img_save > 1.0] = 1.0
        if self.channel_num == 1:
            img_save = Image.fromarray(np.uint8(img_save[...,0] * 255.0 + 0.5))
        else:
            img_save = Image.fromarray(np.uint8(img_save * 255.0 + 0.5))
        img_save.save(str(self.coded_img_dir) + "/coded_img.png")

        print("Saved single coded img in {}".format(self.coded_img_dir))


class NeRFbasedRendering():
    def __init__(self, args, save_dir, device):
        self.save_base = save_dir
        self.out_depth = args.out_depth
        self.device = device
        self.cut_edge = args.cut_edge
        self.channel_num = args.channel_num

        self.pix_mask_size = args.pix_mask_size
        self.pix_mask_num = self.pix_mask_size*self.pix_mask_size
        self.view_point_num = args.angRes*args.angRes

        if self.cut_edge:
            self.edge_size = 9
            
        self.rendered_dir = self.make_savedir("rendered_views")
        if self.out_depth:
            self.depth_dir = self.make_savedir("rendered_depths")

    def make_savedir(self, name):
        savedir = self.save_base.joinpath(name)
        savedir.mkdir(exist_ok=True)
        return savedir

    def render_imgs(self, coded_img, PSreconstruct, vdsr, args, **kwargs):


        def make_LFcoords_test_light(H, W, u_dim, v_dim, u_dim_num, v_dim_num, v_index):
            
            x = np.linspace(0, W-1, W).astype(np.float32)
            y = np.linspace(0, H-1, H).astype(np.float32)
            u = np.linspace(0, u_dim-1, u_dim_num).astype(np.float32)
            v = np.linspace(0, v_dim-1, v_dim_num).astype(np.float32)
            i = [u, v] # same as dataset load (v, u, y, x)
            vv, uv = np.meshgrid(*i)
            uv_grid = np.stack([uv,vv], axis=-1)
            uv_grid = uv_grid[v_index]
            uv_grid = uv_grid[:,None,None,:]
            uv_grid = np.tile(uv_grid, (1,H,W,1))
            i = [x,y]
            yv,xv = np.meshgrid(*i)
            img_grid = np.stack([xv,yv],axis=-1)
            img_grid = img_grid[None,:,:,:]
            img_grid = np.tile(img_grid, (u_dim_num, 1, 1, 1))
            img_grid = np.concatenate([uv_grid, img_grid], -1)
            # img_grid = np.stack([uv, vv, yv, xv], axis=-1)
            
            # set center to 0
            if u_dim % 2 != 0:
                img_grid[...,:2] = img_grid[...,:2] - (u_dim//2)
            elif u_dim % 2 == 0:
                img_grid[...,:2] = img_grid[...,:2] - (u_dim//2 - 0.5)
                
            img_grid[...,2] = img_grid[...,2] - (H // 2)
            img_grid[...,3] = img_grid[...,3] - (W // 2)

            img_grid = img_grid.astype(np.float32)


            return img_grid # (U, V, X, Y)

        self.H = coded_img.shape[1]
        self.W = coded_img.shape[2]
        sub_H = int(self.H/self.pix_mask_size)
        sub_W = int(self.W/self.pix_mask_size)

        coded_img_reshape = coded_img.reshape(coded_img.shape[0], sub_H, self.pix_mask_size, sub_W, self.pix_mask_size).permute(0,2,4,1,3).reshape(coded_img.shape[0], self.pix_mask_num, sub_H, sub_W)

        v_dim_res = args.newview_resolution #13
        view_range = args.view_range #7 [-3 , 3]

                    
        bds_dict = {
            "y_size" : (self.H/2),
            "x_size" : (self.W/2),
            "x_norm" : max((self.H/2), (self.W/2)),
            "y_norm" : max((self.H/2), (self.W/2)),
        }
        kwargs.update(bds_dict)

        if self.cut_edge:
            self.H_ = self.H - 2 * self.edge_size
            self.W_ = self.W - 2 * self.edge_size
        else:
            self.H_ = self.H
            self.W_ = self.W

        tmp_re_Volumes = []
        for ch in range(self.channel_num):
            gray_re_Volume = vdsr[0](PSreconstruct[0](coded_img_reshape))
            gray_re_Volume = gray_re_Volume.reshape(1, args.feature_ch, args.feature_z_num, self.H, self.W)
            tmp_re_Volumes.append(gray_re_Volume)

        for v in trange(v_dim_res):
            
            """Make rays to input MLP"""
            coordinates_t = make_LFcoords_test_light(self.H, self.W, view_range, view_range, v_dim_res, v_dim_res, v)
            coordinates_t = torch.from_numpy(coordinates_t).float()
            view_point_num_nv = v_dim_res
            coordinates_t = coordinates_t.reshape((-1, 4))

            if not self.out_depth:
                re_data_nv = np.zeros([view_point_num_nv, self.H, self.W, self.channel_num], np.float32)
                for ch in range(self.channel_num):

                    """NeRF based rendering"""
                    re_light_field = render(coordinates_t, tmp_re_Volumes[ch], chunk=args.chunk, use_viewdirs=args.use_viewdirs, **kwargs)

                    re_light_field_gray = re_light_field[0].reshape(view_point_num_nv, self.H, self.W)
                    re_data_nv[...,ch] = re_light_field_gray.to("cpu").detach().numpy()

                re_data_nv[re_data_nv < 0.0]  = 0.0
                re_data_nv[re_data_nv > 1.0]  = 1.0

            elif self.out_depth:
                kwargs["extract_return"] = True
                re_data_nv = np.zeros([view_point_num_nv, self.H, self.W, self.channel_num], np.float32)
                re_data_nv_depth = np.zeros([view_point_num_nv, self.H, self.W, 1], np.float32)
                for ch in range(self.channel_num):

                    """NeRF based rendering"""
                    re_light_field = render(coordinates_t, tmp_re_Volumes[ch], chunk=args.chunk, use_viewdirs=args.use_viewdirs, **kwargs)

                    re_light_field_gray = re_light_field[0].reshape(view_point_num_nv, self.H, self.W)
                    re_light_field_depth = re_light_field[1].reshape(view_point_num_nv, self.H, self.W)
                    re_data_nv[...,ch] = re_light_field_gray.to("cpu").detach().numpy()
                    re_data_nv_depth[...,ch] = re_light_field_depth.to("cpu").detach().numpy()

                    re_data_nv[re_data_nv < 0.0]  = 0.0
                    re_data_nv[re_data_nv > 1.0]  = 1.0

                    if args.depthMax is None:
                        Max = np.max(re_data_nv_depth)
                    else:
                        Max = args.depthMax
                    if args.depthMin is None:
                        Min = np.min(re_data_nv_depth)
                    else:
                        Min = args.depthMin
                    re_data_nv_depth = (re_data_nv_depth - Min) / (Max - Min)
                    re_data_nv_depth[re_data_nv_depth < 0.0]  = 0.0
                    re_data_nv_depth[re_data_nv_depth > 1.0]  = 1.0

            ''' Save Reconstructed LF '''
            if self.cut_edge:
                re_data_nv = re_data_nv[...,self.edge_size:-self.edge_size,self.edge_size:-self.edge_size,:]
                if self.out_depth:
                    re_data_nv_depth = re_data_nv_depth[...,self.edge_size:-self.edge_size,self.edge_size:-self.edge_size,:]
            re_light_field_nv = np.zeros([view_point_num_nv, self.H_, self.W_, self.channel_num], np.uint8)
            if self.out_depth:
                re_light_field_nv_depth = np.zeros([view_point_num_nv, self.H_, self.W_, 1], np.uint8)


            for t in range(view_point_num_nv):
                re_light_field_nv[t] = np.uint8(re_data_nv[t]*255.0 + 0.5)
                if self.out_depth:
                    re_light_field_nv_depth[t] = np.uint8(re_data_nv_depth[t]*255.0 + 0.5)

                if args.save_image is True:
                    if self.channel_num == 1:
                        img_save_re_nv = Image.fromarray(re_light_field_nv[t,...,0])

                        if self.out_depth:
                            tmp = re_light_field_nv_depth[t,...,0]
                            cmap = plt.get_cmap("seismic")
                            img_save_re_nv_depth = Image.fromarray(np.uint8((cmap(tmp)) * 255.0 + 0.5))
                    else:
                        img_save_re_nv = Image.fromarray(re_light_field_nv[t])
                        img_save_re_nv_depth = Image.fromarray(re_light_field_nv_depth[t])

                    u = int(t%view_point_num_nv)
                    img_save_re_nv.save(str(self.rendered_dir) + "/%02d_%02d.png" % (v, u))
                    if self.out_depth:
                        img_save_re_nv_depth.save(str(self.depth_dir) + "/%02d_%02d.png" % (v, u))
            
        print("Saved rendered images in {}".format(self.rendered_dir))
        if self.out_depth:
            print("Saved rendered depth in {}".format(self.depth_dir))


